/*
 * rijndael-test.c   v2.3   April '2000
 *
 * Optimised ANSI C code
 *
 * authors: v1.0: Paulo Barreto
 *                Antoon Bosselaers
 *          v2.0: Vincent Rijmen, K.U.Leuven
 *          v2.3: Paulo Barreto
 *
 * This code is placed in the public domain.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define INTERMEDIATE_VALUE_KAT	1
#include "rijndael-alg.c"
#include "rijndael-api.c"

#define SUBMITTER "Joan Daemen"

#define DIR_ENCRYPT	0
#define DIR_DECRYPT	1

static void blockPrintN(FILE *fp, const byte *block, int N, const char *tag) {
	int i;
	fprintf (fp, "%s=", tag);
	for (i = 0; i < N; i++) {
		fprintf (fp, "%02X", block[i]);
	}
	fprintf (fp, "\n");
	fflush (fp);
} /* blockPrintN */

static void blockPrint(FILE *fp, const byte *block, const char *tag) {
	blockPrintN(fp, block, 16, tag);
} /* blockPrint */

static void rijndaelVKKAT(FILE *fp, int keyLength) {
	int i, j, r;
	byte block[4*4];
	char keyMaterial[320];
	byte byteVal = (byte)'8';
	rd_keyinstance keyInst;

#ifdef TRACE_KAT_MCT
	printf("Executing Variable-Key KAT (key %d): ", keyLength);
	fflush(stdout);
#endif /* ?TRACE_KAT_MCT */
	fprintf(fp,
		"\n"
		"==========\n"
		"\n"
		"KEYSIZE=%d\n"
		"\n", keyLength);
	fflush(fp);
	memset(block, 0, 16);
	blockPrint(fp, block, "PT");
	memset(keyMaterial, 0, sizeof (keyMaterial));
	memset(keyMaterial, '0', keyLength/4);
	for (i = 0; i < keyLength; i++) {
		keyMaterial[i/4] = byteVal; /* set only the i-th bit of the i-th test key */
		r = rijndael_makekey(&keyInst, RD_KEY_HEX, keyMaterial);
		if (0 != r) {
			fprintf(stderr,"makeKey error %d\n",r);
			exit(-1);
		}
		fprintf(fp, "\nI=%d\n", i+1);
		fprintf(fp, "KEY=%s\n", keyMaterial);
		memset(block, 0, 16);
		r = rijndael_ecb_encrypt(&keyInst, block, block, 16, NULL);
		if (16 != r) {
			fprintf(stderr,"ecb_encrypt() error %d\n",r);
			exit(-1);
		}
		blockPrint(fp, block, "CT");
		/* now check decryption: */
		rijndael_ecb_decrypt(&keyInst, block, block, 16, NULL);
		for (j = 0; j < 16; j++) {
			assert(block[j] == 0);
		}
		/* undo changes for the next iteration: */
		keyMaterial[i/4] = (byte)'0';
		byteVal =
			(byteVal == '8') ?	'4' :
			(byteVal == '4') ?	'2' :
			(byteVal == '2') ?	'1' :
		/*	(byteVal == '1') */	'8';
	}
	assert(byteVal == (byte)'8');

#ifdef TRACE_KAT_MCT
	printf(" done.\n");
#endif /* ?TRACE_KAT_MCT */
} /* rijndaelVKKAT */


static void rijndaelVTKAT(FILE *fp, int keyLength) {
	int i;
	byte block[4*4];
	char keyMaterial[320];
	rd_keyinstance keyInst;

#ifdef TRACE_KAT_MCT
	printf("Executing Variable-Text KAT (key %d): ", keyLength);
	fflush(stdout);
#endif /* ?TRACE_KAT_MCT */
	fprintf(fp,
		"\n"
		"==========\n"
		"\n"
		"KEYSIZE=%d\n"
		"\n", keyLength);
	fflush(fp);
	memset(keyMaterial, 0, sizeof (keyMaterial));
	memset(keyMaterial, '0', keyLength/4);
	rijndael_makekey(&keyInst, RD_KEY_HEX, keyMaterial);
	fprintf(fp, "KEY=%s\n", keyMaterial);
	for (i = 0; i < 128; i++) {
		memset(block, 0, 16);
		block[i/8] |= 1 << (7 - i%8); /* set only the i-th bit of the i-th test block */
		fprintf (fp, "\nI=%d\n", i+1);
		blockPrint(fp, block, "PT");
		rijndael_ecb_encrypt(&keyInst, block, block, 16, NULL);
		blockPrint(fp, block, "CT");
	}
#ifdef TRACE_KAT_MCT
	printf(" done.\n");
#endif /* ?TRACE_KAT_MCT */
}


static void rijndaelTKAT(FILE *fp,  int keyLength, FILE *in) {
	int i, j;
	char c;
	unsigned int s;
	byte block[4*4], block2[4*4];
	char keyMaterial[320];
	rd_keyinstance keyInst;

#ifdef TRACE_KAT_MCT
	printf("Executing Tables KAT (key %d): ", keyLength);
	fflush(stdout);
#endif /* ?TRACE_KAT_MCT */
	fprintf(fp,
		"\n"
		"==========\n"
		"\n"
		"KEYSIZE=%d\n"
		"\n", keyLength);
	fflush(fp);

	memset(keyMaterial, 0, sizeof (keyMaterial));

	for (i = 0; i < 64; i++) {
		fprintf(fp, "\nI=%d\n", i+1);
		for(j = 0; j < keyLength/4; j++) {
			fscanf(in, "%c", &keyMaterial[j]);
		}
		rijndael_makekey(&keyInst, RD_KEY_HEX, keyMaterial);

		fprintf(fp, "KEY=%s\n", keyMaterial);

		for (j = 0; j < 16; j++) {
			fscanf(in,"%02x",&s);
			block[j] = s;
		}
		fscanf(in, "%c", &c);
		fscanf(in, "%c", &c);
		blockPrint(fp, block, "PT");
		rijndael_ecb_encrypt(&keyInst, block, block2, 16, NULL);
		blockPrint(fp, block2, "CT");
	}
	for (i = 64; i < 128; i++) {
		fprintf(fp, "\nI=%d\n", i+1);
		for(j = 0; j < keyLength/4; j++) {
			fscanf(in, "%c", &keyMaterial[j]);
		}
		rijndael_makekey(&keyInst, RD_KEY_HEX, keyMaterial);

		fprintf(fp, "KEY=%s\n", keyMaterial);

		for (j = 0; j < 16; j++) {
			fscanf(in, "%02x", &s);
			block[j] = s;
		}
		fscanf(in, "%c", &c);
		fscanf(in, "%c", &c);
		rijndael_ecb_decrypt(&keyInst, block, block2, 16, NULL);
		blockPrint(fp, block2, "PT");
		blockPrint(fp, block, "CT");
	}

#ifdef TRACE_KAT_MCT
	printf(" done.\n");
#endif /* ?TRACE_KAT_MCT */
}

static void rijndaelIVKAT (FILE *fp, int keyLength)
{
	int i;
	byte block[4*4], block2[4*4];
	char keyMaterial[320];
	rd_keyinstance keyInst;
	char format[10];
#ifdef TRACE_KAT_MCT
	printf ("Executing Intermediate value KAT (key %d): ", keyLength);
	fflush (stdout);
#endif /* ?TRACE_KAT_MCT */

	fprintf(fp,
		"\n"
		"==========\n"
		"\n"
		"KEYSIZE=%d\n",
		keyLength);
	fflush(fp);
	memset(keyMaterial, 0, sizeof (keyMaterial));
	for (i = 0; i < keyLength/8; i++) {
		sprintf(&keyMaterial[2*i], "%02X", i);
	}
	rijndael_makekey(&keyInst, RD_KEY_HEX, keyMaterial);
	fprintf(fp, "KEY=%s\n", keyMaterial);
	fprintf(fp, "\nIntermediate Ciphertext Values (Encryption)\n\n");
	for (i = 0; i < 16; i++) {
		block[i] = i;
	}
	blockPrint(fp, block, "PT");
	for(i = 1; i < keyInst.rounds; i++) {
		cipherEncryptUpdateRounds(&keyInst, block, block2, i);
		sprintf(format, "CT%d", i);
		blockPrint(fp, block2, format);
	}
	cipherEncryptUpdateRounds(&keyInst, block, block2, keyInst.rounds);
	blockPrint(fp, block2, "CT");

	/* BEGIN HACK */
	/* in order to get the same intermediate values in decryption
	   as in encryption, we need to decrypt using the encryption
	   roundkeys

	   consider it as a dirty trick
	 */
	rijndael_makekey(&keyInst, RD_KEY_HEX, keyMaterial);
	memcpy(keyInst.decsched, keyInst.encsched, sizeof(keyInst.decsched));
	/* END HACK */

	fprintf(fp, "\nIntermediate Ciphertext Values (Decryption)\n\n");
	blockPrint(fp, block2, "CT");
	for(i = 1; i < keyInst.rounds; i++) {
		cipherDecryptUpdateRounds(&keyInst, block2, block, keyInst.rounds - i);
		sprintf(format, "PT%d", i);
		blockPrint(fp, block, format);
	}
	cipherDecryptUpdateRounds(&keyInst, block2, block, 0);
	blockPrint(fp, block, "PT");

#ifdef TRACE_KAT_MCT
	printf(" done.\n");
#endif
}

static void makeKATs (const char *vkFile, const char *vtFile, const char *tblFile, const char *ivFile) {
	FILE *fp, *fp2;

	/* prepare Variable Key Known Answer Tests: */
	fp = fopen(vkFile, "w");
	fprintf(fp,
		"\n"
		"=========================\n"
		"\n"
		"FILENAME:  \"%s\"\n"
		"\n"
		"Electronic Codebook (ECB) Mode\n"
		"Variable Key Known Answer Tests\n"
		"\n"
		"Algorithm Name: Rijndael\n"
		"Principal Submitter: %s\n",
		vkFile, SUBMITTER);
	fflush(fp);

	rijndaelVKKAT(fp, 128); /* test for 128-bit key */
	rijndaelVKKAT(fp, 192); /* test for 192-bit key */
	rijndaelVKKAT(fp, 256); /* test for 256-bit key */

	fprintf(fp,
		"\n"
		"==========");
	fclose(fp);

	/* prepare Variable Text Known Answer Tests: */
	fp = fopen(vtFile, "w");
	fprintf(fp,
		"\n"
		"=========================\n"
		"\n"
		"FILENAME:  \"%s\"\n"
		"\n"
		"Electronic Codebook (ECB) Mode\n"
		"Variable Text Known Answer Tests\n"
		"\n"
		"Algorithm Name: Rijndael\n"
		"Principal Submitter: %s\n",
		vtFile, SUBMITTER);
	fflush(fp);

	rijndaelVTKAT(fp, 128);
	rijndaelVTKAT(fp, 192);
	rijndaelVTKAT(fp, 256);

	fprintf(fp,
		"\n"
		"==========");
	fclose(fp);

	/* prepare Tables Known Answer Tests: */
	fp = fopen(tblFile, "w");
	fprintf(fp,
	    "/* Description of what tables are tested:\n"
		"   The provided implementations each use a different set of tables\n"
		"    - Java implementation: uses no tables\n"
		"    - reference C implementation: uses Logtable, Alogtable, S, Si, rcon\n"
		"    - fast C implementation: uses Logtable, Alogtable,  rcon\n"
		"        and additionally, T1, T2, T3, T4, T5, T6, T7, T8\n"
		"   All these tables are tested.\n"
		"\n"
		"=========================\n"
		"\n"
		"FILENAME:  \"%s\"\n"
		"\n"
		"Electronic Codebook (ECB) Mode\n"
		"Tables Known Answer Tests\n"
		"\n"
		"Algorithm Name: Rijndael\n"
		"Principal Submitter: %s\n",
		tblFile, SUBMITTER);
	fflush(fp);

	if (NULL != (fp2 = fopen("table.128", "r"))) {
		rijndaelTKAT(fp, 128, fp2);
		fclose(fp2);
	} else {
		printf("Table Known Answer test expects file table.128\n");
		fclose(fp);
		exit(EXIT_FAILURE);
	}
	if (NULL != (fp2 = fopen("table.192", "r"))) {
		rijndaelTKAT(fp, 192, fp2);
		fclose(fp2);
	} else {
		printf("Table Known Answer test expects file table.192\n");
		fclose(fp);
		exit(EXIT_FAILURE);
	}
	if (NULL != (fp2 = fopen("table.256", "r"))) {
		rijndaelTKAT (fp, 256, fp2);
		fclose(fp2);
	} else {
		printf("Table Known Answer test expects file table.192\n");
		fclose(fp);
		exit(EXIT_FAILURE);
	}

	fprintf(fp,
		"\n"
		"==========");
	fclose(fp);

	/* prepare Intermediate Values Known Answer Tests: */
	fp = fopen(ivFile, "w");
	fprintf(fp,
		"\n"
		"=========================\n"
		"\n"
		"FILENAME:  \"%s\"\n"
		"\n"
		"Electronic Codebook (ECB) Mode\n"
		"Intermediate Value Known Answer Tests\n"
		"\n"
		"Algorithm Name: Rijndael\n"
		"Principal Submitter: %s\n",
		ivFile, SUBMITTER);
	fflush(fp);

	rijndaelIVKAT(fp, 128);
	rijndaelIVKAT(fp, 192);
	rijndaelIVKAT(fp, 256);

	fprintf(fp,
		"\n"
		"==========");
	fclose(fp);
}

static void rijndaelECB_MCT(FILE *fp, int keyLength, byte direction) {
	int i, j;
	byte inBlock[4*4], outBlock[4*4], binKey[4*MAXKC];
	char keyMaterial[320];
	rd_keyinstance keyInst;

#ifdef TRACE_KAT_MCT
	int width = 0;
	clock_t elapsed = -clock();
	printf("Executing ECB MCT (%s, key %d): ",
		direction == DIR_ENCRYPT ? "ENCRYPT" : "DECRYPT", keyLength);
	fflush(stdout);
#endif /* ?TRACE_KAT_MCT */
	fprintf(fp,
		"\n"
		"=========================\n"
		"\n"
		"KEYSIZE=%d\n", keyLength);
	fflush(fp);
	memset(outBlock, 0, 16);
	memset(binKey, 0, keyLength/8);
	for (i = 0; i < 400; i++) {
#ifdef TRACE_KAT_MCT
        while (width-- > 0) {
        	putchar('\b');
        }
        width = printf("%d", i);
        fflush(stdout);
#endif /* ?TRACE_KAT_MCT */
		fprintf(fp, "\nI=%d\n", i);
		/* prepare key: */
		for (j = 0; j < keyLength/8; j++) {
			sprintf(&keyMaterial[2*j], "%02X", binKey[j]);
		}
		keyMaterial[keyLength/4] = 0;
		fprintf(fp, "KEY=%s\n", keyMaterial);
		rijndael_makekey(&keyInst, /*direction,*/ RD_KEY_HEX, keyMaterial);
		/* do encryption/decryption: */
		blockPrint(fp, outBlock, direction == DIR_ENCRYPT ? "PT" : "CT");
		if (direction == DIR_ENCRYPT) {
			for (j = 0; j < 10000; j++) {
				memcpy(inBlock, outBlock, 16);
				rijndael_ecb_encrypt(&keyInst, inBlock, outBlock, 16, NULL);
			}
		} else {
			for (j = 0; j < 10000; j++) {
				memcpy(inBlock, outBlock, 16);
				rijndael_ecb_decrypt(&keyInst, inBlock, outBlock, 16, NULL);
			}
		}
		blockPrint(fp, outBlock, direction == DIR_ENCRYPT ? "CT" : "PT");
		/* prepare new key: */
		switch (keyLength) {
		case 128:
			for (j = 0; j < 128/8; j++) {
				binKey[j] ^= outBlock[j];
			}
			break;
		case 192:
			for (j = 0; j < 64/8; j++) {
				binKey[j] ^= inBlock[j + 64/8];
			}
			for (j = 0; j < 128/8; j++) {
				binKey[j + 64/8] ^= outBlock[j];
			}
			break;
		case 256:
			for (j = 0; j < 128/8; j++) {
				binKey[j] ^= inBlock[j];
			}
			for (j = 0; j < 128/8; j++) {
				binKey[j + 128/8] ^= outBlock[j];
			}
			break;
		}
	}
#ifdef TRACE_KAT_MCT
	elapsed += clock();
	printf(" done (%.1f s).\n", (float)elapsed/CLOCKS_PER_SEC);
#endif /* ?TRACE_KAT_MCT */
} /* rijndaelECB_MCT */


static void rijndaelCBC_MCT(FILE *fp, int keyLength, byte direction) {
	int i, j, r, t;
	byte inBlock[256/8], outBlock[256/8], binKey[256/8], cv[256/8];
	char keyMaterial[320];
	rd_keyinstance keyInst;

#ifdef TRACE_KAT_MCT
	int width = 0;
	clock_t elapsed = -clock();
	printf("Executing CBC MCT (%s, key %d): ",
		direction == DIR_ENCRYPT ? "ENCRYPT" : "DECRYPT", keyLength);
	fflush (stdout);
#endif /* ?TRACE_KAT_MCT */
	fprintf (fp,
		"\n"
		"==========\n"
		"\n"
		"KEYSIZE=%d\n", keyLength);
	fflush(fp);
	memset(cv,0,16);
	memset(inBlock,0,16);
	memset(binKey,0,keyLength/8);
	for (i = 0; i < 400; i++) {
#ifdef TRACE_KAT_MCT
        while (width-- > 0) {
        	putchar('\b');
        }
        width = printf("%d", i);
        fflush(stdout);
#endif /* ?TRACE_KAT_MCT */
		fprintf (fp, "\nI=%d\n", i);
		/* prepare key: */
		for (j = 0; j < keyLength/8; j++) {
			sprintf (&keyMaterial[2*j], "%02X", binKey[j]);
		}
		keyMaterial[keyLength/4] = 0;
		fprintf(fp, "KEY=%s\n", keyMaterial);
		r = rijndael_makekey(&keyInst, /*direction,*/ RD_KEY_HEX, keyMaterial);
		if (0 != r) {
			fprintf(stderr,"makeKey error %d\n",r);
			exit(-1);
		}
		/* do encryption/decryption: */
		blockPrint(fp, cv, "IV");
		blockPrint(fp, inBlock, direction == DIR_ENCRYPT ? "PT" : "CT");
		if (direction == DIR_ENCRYPT) {
			for (j = 0; j < 10000; j++) {
				for (t = 0; t < 16; t++) {
					inBlock[t] ^= cv[t];
				}
				r = rijndael_ecb_encrypt(&keyInst, inBlock, outBlock, 16, NULL);
				if (16 != r) {
					fprintf(stderr,"ecb_encrypt() error %d\n",r);
					exit(-1);
				}
				memcpy(inBlock, cv, 16);
				memcpy(cv, outBlock, 16);
			}
		} else {
			for (j = 0; j < 10000; j++) {
				rijndael_ecb_decrypt(&keyInst, inBlock, outBlock, 16, NULL);
				for (t = 0; t < 16; t++) {
					outBlock[t] ^= cv[t];
				}
				memcpy(cv, inBlock, 16);
				memcpy(inBlock, outBlock, 16);
			}
		}
		blockPrint(fp, outBlock, direction == DIR_ENCRYPT ? "CT" : "PT");
		/* prepare new key: */
		switch (keyLength) {
		case 128:
			for (j = 0; j < 128/8; j++) {
				binKey[j] ^= outBlock[j];
			}
			break;
		case 192:
			for (j = 0; j < 64/8; j++) {
				if (direction == DIR_ENCRYPT) {
					binKey[j] ^= inBlock[j + 64/8];
				} else {
					binKey[j] ^= cv[j + 64/8];
				}
			}
			for (j = 0; j < 128/8; j++) {
				binKey[j + 64/8] ^= outBlock[j];
			}
			break;
		case 256:
			for (j = 0; j < 128/8; j++) {
				if (direction == DIR_ENCRYPT) {
					binKey[j] ^= inBlock[j];
				} else {
					binKey[j] ^= cv[j];
				}
			}
			for (j = 0; j < 128/8; j++) {
				binKey[j + 128/8] ^= outBlock[j];
			}
			break;
		}
	}
#ifdef TRACE_KAT_MCT
	elapsed += clock();
	printf(" done (%.1f s).\n", (float)elapsed/CLOCKS_PER_SEC);
#endif /* ?TRACE_KAT_MCT */
} /* rijndaelCBC_MCT */


static void makeMCTs(const char *ecbEncryptionFile, const char *ecbDecryptionFile,
		const char *cbcEncryptionFile, const char *cbcDecryptionFile) {
	FILE *fp;

	/* prepare ECB Encryption Monte Carlo Tests: */
	fp = fopen(ecbEncryptionFile, "w");
	fprintf(fp,
		"\n"
		"=========================\n"
		"\n"
		"FILENAME:  \"%s\"\n"
		"\n"
		"Electronic Codebook (ECB) Mode - ENCRYPTION\n"
		"Monte Carlo Test\n"
		"\n"
		"Algorithm Name: Rijndael\n"
		"Principal Submitter: %s\n",
		ecbEncryptionFile, SUBMITTER);
	fflush(fp);

	rijndaelECB_MCT(fp, 128, DIR_ENCRYPT);
	rijndaelECB_MCT(fp, 192, DIR_ENCRYPT);
	rijndaelECB_MCT(fp, 256, DIR_ENCRYPT);

	fprintf(fp,
		"\n"
		"===========");
	fclose(fp);

	/* prepare ECB Decryption Monte Carlo Tests: */
	fp = fopen(ecbDecryptionFile, "w");
	fprintf(fp,
		"\n"
		"=========================\n"
		"\n"
		"FILENAME:  \"%s\"\n"
		"\n"
		"Electronic Codebook (ECB) Mode - DECRYPTION\n"
		"Monte Carlo Test\n"
		"\n"
		"Algorithm Name: Rijndael\n"
		"Principal Submitter: %s\n",
		ecbDecryptionFile, SUBMITTER);
	fflush(fp);

	rijndaelECB_MCT(fp, 128, DIR_DECRYPT);
	rijndaelECB_MCT(fp, 192, DIR_DECRYPT);
	rijndaelECB_MCT(fp, 256, DIR_DECRYPT);

	fprintf(fp,
		"\n"
		"===========");
	fclose(fp);

	/* prepare CBC Encryption Monte Carlo Tests: */
	fp = fopen (cbcEncryptionFile, "w");
	fprintf(fp,
		"\n"
		"=========================\n"
		"\n"
		"FILENAME:  \"%s\"\n"
		"\n"
		"Cipher Block Chaining (CBC) Mode - ENCRYPTION\n"
		"Monte Carlo Test\n"
		"\n"
		"Algorithm Name: Rijndael\n"
		"Principal Submitter: %s\n",
		cbcEncryptionFile, SUBMITTER);
	fflush(fp);

	rijndaelCBC_MCT(fp, 128, DIR_ENCRYPT);
	rijndaelCBC_MCT(fp, 192, DIR_ENCRYPT);
	rijndaelCBC_MCT(fp, 256, DIR_ENCRYPT);


	fprintf(fp,
		"\n"
		"===========");
	fclose(fp);

	/* prepare CBC Decryption Monte Carlo Tests: */
	fp = fopen(cbcDecryptionFile, "w");
	fprintf(fp,
		"\n"
		"=========================\n"
		"\n"
		"FILENAME:  \"%s\"\n"
		"\n"
		"Cipher Block Chaining (CBC) Mode - DECRYPTION\n"
		"Monte Carlo Test\n"
		"\n"
		"Algorithm Name: Rijndael\n"
		"Principal Submitter: %s\n",
		cbcDecryptionFile, SUBMITTER);
	fflush(fp);

	rijndaelCBC_MCT(fp, 128, DIR_DECRYPT);
	rijndaelCBC_MCT(fp, 192, DIR_DECRYPT);
	rijndaelCBC_MCT(fp, 256, DIR_DECRYPT);

	fprintf(fp,
		"\n"
		"===========");
	fclose(fp);
} /* makeMCTs */

void rand_init(void) {
	srand(time(NULL));
}

word32 rand_word32(void) {
	return
		((word32)(rand() >> 7) << 24) |
		((word32)(rand() >> 7) << 16) |
		((word32)(rand() >> 7) <<  8) |
		((word32)(rand() >> 7));
}

static void rijndaelSpeed(FILE *fp, int keyLength,
		rd_function encrypt, rd_function decrypt, int iterations) {
	int i, rounds;
	byte block[4*4];
	char keyMaterial[320];
	byte IV[4*4];
	rd_keyinstance keyInst;
	clock_t elapsed;
	double sec;

	printf("Measuring speed (key %d): \n", keyLength);
	fflush(stdout);

	rounds = iterations;
	rand_init();
	for (i = 3; i >= 0; i--) {
		((word32*)block)[i] = rand_word32();
	}
   	blockPrint(fp, block, "plaintext");

#if 0	/* How could this have worked?  -- kjb */
	memset(keyMaterial, 0, sizeof (keyMaterial));
	memset(keyMaterial, '0', keyLength/4);
	for (i = 0; i < keyLength/4; i++) {
		keyMaterial[i] = '0'+ i;
	}
#endif
	for (i = 0; i < keyLength/4; i++) {
		sprintf(&keyMaterial[i], "%X", i & 0xF);
	}

	printf("Measuring key scheduling speed...");
	elapsed = -clock ();
	for (i = rounds; i > 0; i--) {
		rijndael_makekey(&keyInst, RD_KEY_HEX, keyMaterial);
	}
	elapsed += clock ();
	sec = elapsed ? (double)elapsed/CLOCKS_PER_SEC : 1.0;
	printf(" %.2f sec, %.1f Mbit/sec.\n",
		sec, (double)keyLength*iterations/1E6/sec);

	printf("Measuring encryption speed...");
	elapsed = -clock ();
	for (i = rounds; i > 0; i--) {
		if (encrypt != rijndael_ecb_encrypt) memset(IV, 0xA5, sizeof(IV));
		(*encrypt)(&keyInst, block, block, 16, IV);
	}
	elapsed += clock ();

	sec = elapsed ? (double)elapsed/CLOCKS_PER_SEC : 1.0;
	printf(" %.2f sec, %.1f Mbit/sec.\n",
		sec, (double)128*iterations/1E6/sec);
   	blockPrint (fp, block, "encrypted text");

	printf("Measuring decryption speed...");
	elapsed = -clock ();
	for (i = rounds; i > 0; i--) {
		if (decrypt != rijndael_ecb_decrypt) memset(IV, 0xA5, sizeof(IV));
		(*decrypt)(&keyInst, block, block, 16, IV);
	}
	elapsed += clock ();
	sec = elapsed ? (double)elapsed/CLOCKS_PER_SEC : 1.0;
	printf(" %.2f sec, %.1f Mbit/sec.\n",
		sec, (double)128*iterations/1E6/sec);
   	blockPrint(fp, block, "decrypted text");

	printf("done.\n");
} /* rijndaelSpeed */

static void oddtest(FILE *fp, char *name, rd_function encrypt, rd_function decrypt) {
	/* Encrypt and decrypt odd length arrays to test the cypher stealing,
	 * IV propagation and misalignment.  Input and output buffers are not
	 * the same to test if data is copied properly.
	 */
	byte keyMaterial[16];
	rd_keyinstance keyInst;
	byte plain1[2*3*16], crypt[2*3*16], plain2[2*3*16];
	byte IV0[16], iv[16];
	int i;

	rand_init();
	for (i = 3; i >= 0; i--) {
		((word32*)keyMaterial)[i] = rand_word32();
		((word32*)IV0)[i] = rand_word32();
	}
	for (i = 2*3*16/4; i >= 0; i--) {
		((word32*)plain1)[i] = rand_word32();
	}
   	rijndael_makekey(&keyInst, 16, keyMaterial);

   	for (i = 16; i <= 3*16; i++) {
		memset(crypt, 0, sizeof(crypt));
		memset(plain2, 0, sizeof(plain2));

		memcpy(iv, IV0, sizeof(iv));
		(*encrypt)(&keyInst, plain1 + 0, crypt + 0, i, &iv);
		(*encrypt)(&keyInst, plain1 + i, crypt + i, i, &iv);

		memcpy(iv, IV0, sizeof(iv));
		(*decrypt)(&keyInst, crypt + 0, plain2 + 0, i, &iv);
		(*decrypt)(&keyInst, crypt + i, plain2 + i, i, &iv);

		fprintf(fp, "%s: 2 times %d (%s):\n", name, i,
			memcmp(plain1, plain2, i+i) == 0 ? "OK" : "FAIL");
		blockPrintN(fp, plain1, i+i, "plain  ");
		blockPrintN(fp, crypt, i+i, "encrypt");
		blockPrintN(fp, plain2, i+i, "decrypt");
	}
} /* oddtest */

int main(int argc, char **argv) {
	int i;

	i = 1;
	do {
		if (i < argc && strcmp(argv[i], "kat") == 0) {
		} else
		if (i < argc && strcmp(argv[i], "speed") == 0) {
		} else
		if (i < argc && strcmp(argv[i], "odd") == 0) {
		} else {
			fprintf(stderr, "Usage: %s [kat] [speed] [odd]...\n", argv[0]);
			exit(1);
		}
	} while (++i < argc);

	for (i= 1; i < argc; i++) switch (argv[i][0]) {
	case 'k':
		makeKATs("ecb_vk.txt", "ecb_vt.txt", "ecb_tbl.txt", "ecb_iv.txt");
		makeMCTs("ecb_e_m.txt", "ecb_d_m.txt", "cbc_e_m.txt", "cbc_d_m.txt");
		break;
	case 's':
		printf("ECB:\n");
		rijndaelSpeed(stdout, 128, rijndael_ecb_encrypt, rijndael_ecb_decrypt, 1000000);
		rijndaelSpeed(stdout, 192, rijndael_ecb_encrypt, rijndael_ecb_decrypt, 1000000);
		rijndaelSpeed(stdout, 256, rijndael_ecb_encrypt, rijndael_ecb_decrypt, 1000000);
		printf("CFB1:\n");
		rijndaelSpeed(stdout, 128, rijndael_cfb1_encrypt, rijndael_cfb1_decrypt, 10000);
		break;
	case 'o':
		oddtest(stdout, "ECB", rijndael_ecb_encrypt, rijndael_ecb_decrypt);
		oddtest(stdout, "CBC", rijndael_cbc_encrypt, rijndael_cbc_decrypt);
		oddtest(stdout, "CFB1", rijndael_cfb1_encrypt, rijndael_cfb1_decrypt);
		oddtest(stdout, "CFB8", rijndael_cfb8_encrypt, rijndael_cfb8_decrypt);
		break;
	}
	return 0;
}
